/*
 * grammar - Read YACC Grammar
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#if HAS_STDLIB
# include <stdlib.h>
#endif /* HAS_STDLIB */

#include "common.h"
#include "yacc.h"
#include "misc.h"
#include "token.h"
#include "grammar.h"

#ifndef global

#define MAXSYM (MAXNONT + MAXTERM)

#if MAXSYM < 256
typedef unsigned char Gsym;
#else
typedef short Gsym;
#endif

typedef Gsym *Item;

#define A_UNDEF	0
#define A_LEFT	1
#define A_RIGHT	2
#define A_NON	3
#define A_MASK	3

#define NILSYM MAXSYM
#define NB MAXTERM

typedef struct symbol {
    struct symbol *next;
    struct symbol *type;
    short value;
    Gsym code;
    uchar prec;
    uchar assoc;
#define F_EMPTY 0x80
#define F_CONTEXT 0x40
    char name[1];
} Symbol;

typedef struct production {
    short link; /* next prod# which has same LHS */
    uchar assoc;
    uchar prec;
    Gsym body[1];
} Production;

/* Return non-zero if grammar symbol s is a terminal */
#define isterm(s) ((s) < MAXTERM)

#endif /* global */

global Symbol *gsym[MAXSYM + 1];

global Production *gram[MAXPROD];
global int nprods, nterms, nnonts;


global int worst_error;



#define HASHSIZE 256
static Symbol *hashtbl[HASHSIZE];

static char tokbuf[100];
static int cur_prec;
static Gsym start_prime, start_sym;

static bool unioned;

static char M_SYNTAX[] = "Syntax error near '%s'";





/* Insert #line into the parser code */
global void insertline(fp)
    FILE *fp;
{
    if (!lflag)
	fprintf(fp, "#line %d \"%s\"\n", lineno, filename);
}


/* Return string p's hash value */
unsigned hash(p)
    char *p;
{
    unsigned u;
    
    u = 0;
    while (*p)
	u = u * 257 + *p++;
    return (u);
}


/* Return a symbol which corresponds to string s */
Symbol *intern(s)
    char *s;
{
    Symbol *p, **root;
    
    root = hashtbl + (hash(s) % HASHSIZE);
    for (p = *root; p != NULL; p = p->next) {
	if (strcmp(p->name, s) == 0)
	    return (p);
    }
    p = alloc(sizeof(Symbol) + strlen(s));
    p->next = *root;
    *root = p;
    strcpy(p->name, s);
    p->code = NILSYM;
    return (p);
}


/* Return non-zero if t is a terminal or a nonterminal symbol */
#define isgsym(t) ((t) == NAME || (t) == '\'')


/* Return non-zero if c is a octal digit */
#define isoctal(c) ('0' <= (c) && (c) <= '7')


/* Return value of C character sequence s */
int charval(s)
    char *s;
{
    int val;
    int c;
    int i;
    
    if ((c = *s++) != '\\')
	return c;
    c = *s++;
    if (isoctal(c)) {
	val = 0;
	for (i = 0; isoctal(*s) && i < 3; i++)
	    val = val * 8 + *s++ - '0';
	return val;
    }
    switch(c) {
    case 'n': return '\n';
    case 't': return '\t';
    case 'b': return '\b';
    case 'r': return '\r';
    case 'f': return '\f';
    case 'v': return '\013';
    case 'a': return '\007';
    default:
	error("Undefined escape sequence");
	worst_error = 1;
	 /* fall thru */
    case '\'':
    case '\"':
    case '\\':
    case '\?':
	 return c;
     }
}


/* Return grammar id for string s, register if new */
Gsym intern_gsym(s, term)
    char *s;
    bool term;
{
    Symbol *p;
    
    p = intern(s);
    if (p->code != NILSYM)
	return (p->code);
    if (term || p->name[0] == '\'') {
	if (nterms >= MAXTERM)
	    die("Too many terminals");
	gsym[nterms] = p;
	p->code = nterms++;
	p->assoc = A_UNDEF;
	p->prec = 0;
	p->type = NULL;
	if (p->name[0] == '\'') {
	    p->value = charval(p->name + 1);
	} else {
	    p->value = -1;
	}
    } else {
	if (nnonts >= MAXNONT)
	    die("Too many nonterminals");
	gsym[NB + nnonts] = p;
	p->code = NB + nnonts++;
	p->assoc = A_UNDEF;
	p->prec = 0;
	p->type = NULL;
	p->value = -1;
    }
    return (p->code);
}


/* Return unique nonterminal id */
Gsym gen_nont()
{
    static int n = 1;
    char buf[6];
    
    sprintf(buf, "@%d", n++);
    return (intern_gsym(buf, NO));
}


/* Read <typename> in %type/%token, etc. */
Symbol *gettype()
{
    Symbol *type;
    
    if (gettoken(tokbuf) != '<') {
	ungettok(tokbuf);
	return (NULL);
    }
    if (gettoken(tokbuf) != NAME)
	die("Identifier expected");
    type = intern(tokbuf);
    if (gettoken(tokbuf) != '>')
	die("> expected");
    unioned = YES;
    return (type);
}


/* Process %token, %left, %right, %nonassoc */
void do_token(tag)
    int tag;
{
    Symbol *type, *p;
    int precincr = 0;
    char t;
    
    type = gettype();
    t = gettoken(tokbuf);
    while (isgsym(t)) {
	p = gsym[intern_gsym(tokbuf, YES)];
	if (type != NULL)
	    p->type = type;
	switch (tag) {
	case LEFT:
	    p->assoc |= A_LEFT;
	    goto setprec;
	    break;
	case RIGHT:
	    p->assoc |= A_RIGHT;
	    goto setprec;
	case NONASSOC:
	    p->assoc |= A_NON;
	setprec:
	    p->prec = cur_prec;
	    precincr = 1;
	    break;
	}
	if ((t = gettoken(tokbuf)) == NUMBER) {
	    if (p->value == -1)
		p->value = atoi(tokbuf);
	    else
		error1("token %s already has a value", p->name);
	    t = gettoken(tokbuf);
	}
	if (t == ',')
	    t = gettoken(tokbuf);
    }
    ungettok(tokbuf);
    cur_prec += precincr;
}


/* Process %type statement */
void do_type()
{
    Symbol *type, *p;
    char t;
    
    type = gettype();
    for (;;) {
	if ((t = gettoken(tokbuf)) == ',')
	    continue;
	if (!isgsym(t))
	    break;
	p = gsym[intern_gsym(tokbuf, NO)];
	if (type != NULL)
	    p->type = type;
    }
    ungettok(tokbuf);
}


/* Process %union statement */
void do_union()
{
    int ct;
    
    if (gettoken(tokbuf) != '{')
	die("{ expected");
    insertline(ofp);
    fprintf(ofp, "typedef union {");
    if (hfp) fprintf(hfp, "typedef union {");
    ct = 1;
    while (ct != 0) {
	switch (copying_gettoken(tokbuf, ofp, hfp)) {
	case EOF:
	    die("Unexpected EOF");
	    break;
	case '{':
	    ct++;
	    break;
	case '}':
	    ct--;
	    break;
	}
	fputs(tokbuf, ofp);
	if (hfp) fputs(tokbuf, hfp);
    }
    fprintf(ofp, " YYSTYPE;\n");
    if (hfp) fprintf(hfp, " YYSTYPE;\n");
}





/* Copy C text surrounded by %{ ... %} */
void do_copy()
{
    insertline(ofp);
    while (copying_gettoken(tokbuf, ofp, NULL) != ENDINC)
	fputs(tokbuf, ofp);
}



/* Process declaration part */
global void do_declaration()
{
    int t;
    int i;
    int val;

    nterms = nnonts = 0;
    intern_gsym("$EOF", YES);
    gsym[0]->value = 0;
    intern_gsym("error", YES);
    start_prime = intern_gsym("$start", NO);
    start_sym = 0;
    cur_prec = 0;
    unioned = NO;
    while ((t = gettoken(tokbuf)) != MARK) {
	switch (t) {
	case TOKEN:
	case LEFT:
	case RIGHT:
	case NONASSOC:
	    do_token(t);
	    break;
	case BEGININC:
	    do_copy();
	    break;
	case UNION:
	    do_union();
	    unioned = YES;
	    break;
	case TYPE:
	    do_type();
	    break;
	case EXPECT:
	    if (gettoken(tokbuf) == NUMBER) {
		expected_srconf = atoi(tokbuf);
	    } else {
		die("Missing number");
	    }
	    break;
	case START:
	    gettoken(tokbuf);
	    start_sym = intern_gsym(tokbuf, NO);
	    break;
	case EOF:
	    die("No grammar given");
	    break;
	default:
	    die1(M_SYNTAX, tokbuf);
	    break;
	}
    }
    if (!unioned) {
	if (hfp) {
	    fprintf(hfp, "\n");
	    fprintf(hfp, "#ifndef YYSTYPE\n");
	    fprintf(hfp, "#define YYSTYPE int\n");
	    fprintf(hfp, "#endif\n");
	}
	fprintf(ofp, "\n");
	fprintf(ofp, "#ifndef YYSTYPE\n");
	fprintf(ofp, "#define YYSTYPE int\n");
	fprintf(ofp, "#endif\n");
    }
    if (hfp) fprintf(hfp, "extern YYSTYPE %slval;\n", pspref ? pspref : "yy");
    
    /* Assign values to tokens */
    val = gsym[1]->value;
    if (val > 0)
	val++;
    else
	val = 256;
    for (i = 1; i < nterms; i++) {
	if (gsym[i]->value < 0)
	    gsym[i]->value = val++;
    }
    fprintf(ofp, "\n#define YYERRTOK %d\n", gsym[1]->value);
    if (hfp) fprintf(hfp, "\n#define YYERRTOK %d\n", gsym[1]->value);
    for (i = 2; i < nterms; i++) {
	if (gsym[i]->name[0] != '\'') {
	    if (hfp)
		fprintf(hfp, "#define %s %d\n",
			gsym[i]->name, gsym[i]->value);
	    fprintf(ofp, "#define %s %d\n",
		    gsym[i]->name, gsym[i]->value);
	}
    }
}


/* Copy action into C source file */
void copyact(g, n, delm)
    Gsym *g;
    int n;
    int delm;
{
    int c, b;
    int ct, v;
    Symbol *type;
    FILE *fp;
    
    fp = ofp;
    if (afp) {
	/* Actions are in a single switch statement */
	fp = afp;
	fprintf(fp, "case %d:\n", nprods);
    } else {
	/* Actions are independent functions */
	fprintf(fp, "static void yya%d()\n", nprods);
    }
    insertline(fp);
    putc('{', fp);
    ct = 0;
    while ((b = copying_gettoken(tokbuf, fp, NULL)) != delm || ct > 0) {
	switch (b) {
	case EOF:
	    die("Unexpected EOF");
	    break;
	case '{':
	    fputs(tokbuf, fp);
	    ct++;
	    break;
	case '}':
	    fputs(tokbuf, fp);
	    ct--;
	    break;
	case '$':
	    c = get();
	    type = NULL;
	    if (c == '<') {
		gettoken(tokbuf);
		type = intern(tokbuf);
		if (get() != '>')
		    die("> expected");
		c = get();
	    }
	    if (c == '$') {
		fprintf(fp, "yyval");
		v = 0;
	    } else if (c == '-' || isdigit(c)) {
		unget(c);
		v = getnum();
		if (v > n)
		    goto err;
		fprintf(fp, "YYASP(%d)", n - v);
	    } else
		goto err;
	    if (type != NULL ||
		(c == '$' || v > 0) && (type = gsym[g[v]]->type) != NULL)
		fprintf(fp, ".%s", type->name);
	    else if (unioned) {
		error1("type not defined for `%s'", gsym[g[v]]->name);
		worst_error = 1;
	    }
	    break;
	err:
	    error("Bad '$'");
	    worst_error = 1;
	    break;
	default:
	    fputs(tokbuf, fp);
	    break;
	}
    }
    fputs(tokbuf, fp);
    if (delm != '}')
	fprintf(fp, " }");
    fprintf(fp, afp ? " break;\n" : "\n");
}



/* Read grammar */
global void do_grammar()
{
    Gsym gbuf[100], lastterm, w;
    Production *r;
    int i;
    bool act;
    int t;
    
    r = alloc(sizeof(Production) + sizeof(Gsym) * 3);
    r->body[1] = start_prime;
    r->link = -1;
    gram[0] = r;
    nprods = 1;
    gbuf[0] = 0;
    t = gettoken(tokbuf);
    while (t != MARK && t != EOF) {
	if (t == NAME) {
	    gbuf[0] = intern_gsym(tokbuf, NO);
	    if (isterm(gbuf[0]))
		die1("Nonterminal symbol expected: '%s'", tokbuf);
	    if (gettoken(tokbuf) != ':')
		die("':' expected");
	    if (start_sym == 0)
		start_sym = gbuf[0];
	} else if (t == '|') {
	    if (gbuf[0] == 0)
		die1(M_SYNTAX, tokbuf);
	} else if (t == BEGININC) {
	    do_copy();
	    t = gettoken(tokbuf);
	    continue;
	} else {
	    die1(M_SYNTAX, tokbuf);
	}
	lastterm = 0;
	act = NO;
	i = 1;
	for (;;) {
	    t = gettoken(tokbuf);
	    if (t == '=') {
		if ((t = gettoken(tokbuf)) == '{')
		    copyact(gbuf, i - 1, '}');
		else {
		    ungettok(tokbuf);
		    copyact(gbuf, i - 1, ';');
		}
		act = YES;
	    } else if (t == '{') {
		copyact(gbuf, i - 1, '}');
		act = YES;
	    }
	    else if (t == PRECTOK) {
		gettoken(tokbuf);
		lastterm = intern_gsym(tokbuf, NO);
	    }
	    else if (t == NAME && peekch() == ':')
		break;
	    else if (isgsym(t)) {
		if (act) {
		    r = alloc(sizeof(Production) + sizeof(Gsym) * 2);
		    gbuf[i++] = r->body[1] = gen_nont();
		    r->link = gsym[r->body[1]]->value;
		    gsym[r->body[1]]->value = nprods;
		    gram[nprods++] = r;
		}
		gbuf[i++] = w = intern_gsym(tokbuf, NO);
		if (isterm(w))
		    lastterm = w;
		act = NO;
	    }
	    else
		break;
	}
	gbuf[i++] = 0;
	if (!act) {
	    if (i > 2 && gsym[gbuf[0]]->type != NULL
		&& gsym[gbuf[0]]->type != gsym[gbuf[1]]->type) {
		error("Stack types are different");
		worst_error = 1;
	    }
	    if (!afp)
		fprintf(ofp, "#define yya%d yydef\n", nprods);
	}
	r = alloc(sizeof(Production) + sizeof(Gsym) * i);
	memcpy(r->body + 1, gbuf, sizeof(Gsym) * i);
	r->prec = gsym[lastterm]->prec;
	r->assoc = gsym[lastterm]->assoc & A_MASK;
	r->link = gsym[r->body[1]]->value;
	gsym[r->body[1]]->value = nprods;
	gram[nprods++] = r;
	if (t == ';')
	    t = gettoken(tokbuf);
    }
    
    gram[0]->body[2] = start_sym;
    gsym[start_prime]->value = 0;

    for (i = NB + 1; i < NB + nnonts; i++) {
	/* reverse same-LHS link */
	int j, k, w;

	if ((j = gsym[i]->value) == -1) {
	    error1("Nonterminal '%s' used but not defined", gsym[i]->name);
	    worst_error = 1;
	}
	k = -1;
	while (j != -1) {
	    w = gram[j]->link;
	    gram[j]->link = k;
	    k = j;
	    j = w;
	}
	gsym[i]->value = k;
    }

    if (!afp) {
	/* Generate actions */
	fprintf(ofp, "static void (*yya[])() = {\n");
	fprintf(ofp, " 0,");
	for (i = 1; i < nprods; i++) {
	    if (i % 10 == 0)
		fprintf(ofp, "\n");
	    fprintf(ofp, "yya%d,", i);
	}
	fprintf(ofp, "\n};\n\n");
    }

#if 0
    if (debug) {
	Gsym *p;

	fprintf(vfp, "grammar read:\n");
	for (i = 0; i < nprods; i++) {
	    fprintf(vfp, "(%d)\t%s\t:", i, gsym[gram[i]->body[1]]->name);
	    for (p = gram[i]->body + 2; *p != 0; p++)
		fprintf(vfp, " %s", gsym[*p]->name);
	    putc('\n', vfp);
	}
    }
#endif
}
