/*
 * LALR(1) parser generation
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#if HAS_STDLIB
# include <stdlib.h>
#endif /* HAS_STDLIB */
#include <assert.h>

#include "common.h"
#include "yacc.h"
#include "misc.h"
#include "grammar.h"
#include "lalr.h"



#ifndef global

typedef struct {
    Gsym sym;
    short num;
} Reduce;

#endif /* global */

typedef struct {
    Gsym left;
    Item item;
    uchar *look;
} LR1;

global int nstates;
global short *action[MAXSTATE];
global Reduce *reduce[MAXSTATE];
global Gsym thru[MAXSTATE];

static LR1 **kernel;


static int nlooks;
static int nacts, nacts2;
static int nitems;

static int nsrerr = 0;
static int nrrerr = 0;

#if MAXSYM > MAXPROD
static bool tmpflag[MAXSYM];
#else
static bool tmpflag[MAXPROD];
#endif


/* Set-operation functions */
#define NBITS 8 /* number of bits in a byte */

#define tbit(a, i) (((a)[(i) / NBITS] >> ((i) % NBITS)) & 1)
#define sbit(a, i) ((a)[(i) / NBITS] |= 1 << ((i) % NBITS))
#define cbit(a, i) ((a)[(i) / NBITS] &= ~(1 << ((i) % NBITS)))

static bool *nullable;
static uchar (*first)[(MAXTERM + NBITS - 1) / NBITS];
static uchar (*follow)[(MAXTERM + NBITS - 1) / NBITS];



/* Let set d = set d OR set s */
bool orbits(d, s)
    uchar *d;
    uchar *s;
{
    bool changed;
    int i;
    
    changed = NO;
    for (i = 0; i < nterms; i += NBITS) {
	if (*s & ~*d) {
	    changed = YES;
	    *d |= *s;
	}
	s++;
	d++;
    }
    return (changed);
}


#define for_each_member(v, set) \
    for (v = 0; (v = next_elem(v, set)) < nterms; v++)

/* Return an element next to e in set p */
Gsym next_elem(e, p)
    Gsym e;
    uchar *p;
{
    for (; e < nterms; e++) {
	if (tbit(p, e))
	    break;
    }
    return (e);
}


/* Dump set */
void dump_set(set)
    uchar *set;
{
    Gsym x;
    
    for (x = 0; x < nterms; x++) {
	if (tbit(set, x))
	    fprintf(vfp, "%s ", gsym[x]->name);
    }
}


/* Return production number of Item p */
int item2pnum(p)
    Gsym *p;
{
    int i;
    
    while (*--p != 0)
	;
    for (i = 0; i < nprods; i++) {
	if (gram[i]->body == p)
	    return (i);
    }
    die("item2pnum(): can't happen");
}


/* Print Item s */
void write_item(s)
    Item s;
{
    Item p;
    
    p = s;
    while (*--s != 0)
	;
    s++;
    fprintf(vfp, "\t(%d) %s :", item2pnum(p), gsym[*s]->name);
    for (s++; ;s++) {
	if (s == p)
	    fputs(" .", vfp);
	if (*s == '\0')
	    break;
	fprintf(vfp, " %s", gsym[*s]->name);
    }
    fprintf(vfp, "\n");
}




/* Precompute FIRST & NULLABLE for each nonterminals */

    
void first_nullable_precomp()
{
    int i;
    Gsym g, h, *s;
    bool changed;

    nullable = talloc(sizeof(bool) * nnonts);
    first = talloc(sizeof(first[0]) * nnonts);
    do {
	changed = NO;
	for (i = 0; i < nprods; i++) {
	    h = gram[i]->body[1];
	    for (s = gram[i]->body + 2; (g = *s++) != 0; ) {
		if (isterm(g)) {
		    if (!tbit(first[h - NB], g))
			changed = YES;
		    sbit(first[h - NB], g);
		    break;
		}
		changed |= orbits(first[h - NB], first[g - NB]);
		if (!nullable[g - NB])
		    break;
	    }
	    if (g == 0 && !nullable[h - NB]) {
		nullable[h - NB] = YES;
		changed = YES;
	    }
	}
    } while (changed);

    if (debug) {
	fprintf(vfp, "First:\n");
	for (i = 0; i < nnonts; i++) {
	    fprintf(vfp, "%s\t[ ", gsym[NB + i]->name);
	    dump_set(first[i]);
	    if (nullable[i])
		fprintf(vfp, "@ ");
	    fprintf(vfp, "]\n");
	}
    }
}



/* Precompute nonterminals that reaches A -> empty */

void comp_empty()
{
    int i;
    bool changed;

    do {
	changed = 0;
	for (i = 0; i < nprods; i++) {
	    Gsym lhs = gram[i]->body[1];
	    Gsym rhs = gram[i]->body[2];
	    if ((rhs == 0 || gsym[rhs]->assoc & F_EMPTY)
		&& !(gsym[lhs]->assoc & F_EMPTY)) {
		gsym[lhs]->assoc |= F_EMPTY;
		changed = 1;
	    }
	}
    } while (changed);

    if (debug) {
	fprintf(vfp, "EMPTY nonterminals: ");
	for (i = 0; i < nnonts; i++) {
	    if (gsym[NB+i]->assoc & F_EMPTY)
		fprintf(vfp, " %s", gsym[NB+i]->name);
	}
	fprintf(vfp, "\n");
    }
}



/*  compute LALR(1) sets of items  */

typedef struct tmpitem {
    Gsym left;
    Item item;
} Tmpitem;

int bytes;


/* Return non-zero if item set p and t are same */
bool sameset(p, t, n)
    LR1 *p;
    Tmpitem *t;
    int n;
{
    while (n != 0) {
	if (p->item != t->item)
	    return (NO);
	p++;
	t++;
	n--;
    }
    return (p->item == NULL || p->item[-2] == 0);
}

#define MAXEMP 100

/* find empty-productions derived from nonterminal x */
int find_empty(emp, nemp, x)
    Item *emp;
    int nemp;
    Gsym x;
{
    int i;

    if (!tmpflag[x] && gsym[x]->assoc & F_EMPTY) {
	tmpflag[x] = 1;
	for (i = gsym[x]->value; i >= 0; i = gram[i]->link) {
	    if (gram[i]->body[2] == 0) {
		if (nemp >= MAXEMP)
		    die("too many empty productions");
		emp[nemp] = gram[i]->body + 2;
		nemp++;
	    } else if (!isterm(gram[i]->body[2])) {
		nemp = find_empty(emp, nemp, gram[i]->body[2]);
	    }
	}
    }
    return nemp;
}


/* Return LR(1) set of items */
LR1 *make_state(t, n)
    Tmpitem *t;
    int n;
{
    LR1 *stp, *p, *q;
    int i;

    Item emp[MAXEMP];
    int nemp;
    Gsym g;

    /* find empty productions */
    memset(tmpflag, 0, sizeof(bool) * MAXSYM);
    nemp = 0;
    for (i = 0; i < n; i++) {
	g = *t[i].item;
	if (!isterm(g))
	    nemp = find_empty(emp, nemp, g);
    }

    /* copy items */
    stp = talloc(sizeof(LR1) * (n + nemp + 1));
    nitems += (n + nemp);
    p = stp;
    for (i = 0; i < n; i++) {
	p->item = t[i].item;
	if ((p->left = t[i].left) == 0)
	    q = p;
	else
	    for (q = stp; q->left != p->left; q++)
		;
	if (q < p)
	    p->look = q->look;
	else {
	    p->look = talloc(bytes);
	    nlooks++;
	}
	p++;
    }
    /* add empty productions */
    for (i = 0; i < nemp; i++) {
	p->left = 0;
	p->item = emp[i];
	p->look = talloc(bytes);
	p++;
    }
    p->item = NULL;
    return (stp);
}


int cmptemp(x, y)
    Tmpitem *x;
    Tmpitem *y;
{
    int w;
    
    if ((w = (int)x->item[-1] - (int)y->item[-1]) != 0)
	return (w);
    return (x->item - y->item);
}


/* Make LR(0) items (kernel only) */
void comp_kernels()
{
    LR1 *x;
    Gsym *s, g;
    int i, j, k, n, ni;
    int ntemp;
    Tmpitem temp[500];
    short nextst[300];
    
    bytes = (nterms + NBITS - 1) / NBITS;
    temp[0].left = 0;
    temp[0].item = gram[0]->body + 2;
    thru[0] = NILSYM;
    kernel[0] = make_state(temp, 1);
    nstates = 1;
    for (i = 0; i < nstates; i++) {
	/* collect direct GOTO's (come from kernel items) */
	ntemp = 0;
	for (x = kernel[i]; x->item != NULL; x++) {
	    if ((g = *(s = x->item)) != 0) {
		if (ntemp >= height(temp))
		    die("Too many items");
		temp[ntemp].left = 0;
		temp[ntemp].item = s + 1;
		ntemp++;
	    }
	}
	
	/* collect indirect GOTO's (come from nonkernel items) */
	memset(tmpflag, 0, sizeof(bool) * MAXSYM);
	for (j = 0; j < ntemp; j++) {
	    g = temp[j].item[-1];
	    if (!isterm(g) && !tmpflag[g]) {
		tmpflag[g] = YES;
		for (k = gsym[g]->value; k >= 0; k = gram[k]->link) {
		    if (gram[k]->body[2] != 0) {
			if (ntemp >= height(temp))
			    die("Too many items");
			temp[ntemp].left = g;
			temp[ntemp].item = gram[k]->body + 3;
			ntemp++;
		    }
		}
	    }
	}

	qsort(temp, ntemp, sizeof(Tmpitem), cmptemp);
	
	/* compute next states & fill nextst[] */
	ni = 0;
	for (j = 0; j < ntemp; ) {
	    g = temp[j].item[-1];
	    for (k = j; j < ntemp && temp[j].item[-1] == g; j++)
		;
	    for (n = 0; n < nstates; n++) {
		if (sameset(kernel[n], temp + k, j - k))
		    break;
	    }
	    if (n >= nstates) {
		/* brand new state */
		if (nstates >= MAXSTATE)
		    die("Too many states");
		kernel[nstates] = make_state(temp + k, j - k);
		thru[nstates] = g;
		nstates++;
	    }
	    if (ni >= height(nextst))
		die("Too many actions in one state");
	    nextst[ni++] = n;
	}
	action[i] = alloc(sizeof(short) * (ni + 1));
	memcpy(action[i], nextst, sizeof(short) * ni);
	action[i][ni] = 0;
	nacts += ni;
    }
}



/* Compute FIRST set of sequence s into p */
void comp_first(p, s)
    uchar *p;
    Gsym *s;
{
    Gsym g;
    
    while ((g = *s++) != 0) {
	if (isterm(g)) {
	    sbit(p, g);
	    return;
	}
	orbits(p, first[g - NB]);
	if (!nullable[g - NB])
	    return;
    }
}


/* Return non-zero if grammar sequence s could be empty */
bool seq_nullable(s)
    Gsym *s;
{
    Gsym g;
    
    while ((g = *s++) != 0) {
	if (isterm(g) || !nullable[g - NB])
	    return (NO);
    }
    return (YES);
}




/* Compute FOLLOW set into follow */
void comp_follow(st)
    int st;
{
    uchar *p;
    short *t;
    LR1 *x;
    bool changed;
    
    for (t = action[st]; *t; t++) {
	if (!isterm(thru[*t])) {
	    memset(p = follow[thru[*t] - NB], 0, bytes);
	    for (x = kernel[*t]; x->item && x->item[-2]; x++)
		comp_first(p, x->item);
	}
    }
    for (x = kernel[st]; x->item; x++) {
	if (!isterm(*x->item) && seq_nullable(x->item + 1))
	    orbits(follow[*x->item - NB], x->look);
    }
    do {
	changed = NO;
	for (t = action[st]; *t; t++) {
	    if (!isterm(thru[*t])) {
		p = follow[thru[*t] - NB];
		for (x = kernel[*t]; x->item && x->item[-2]; x++) {
		    if (seq_nullable(x->item) && x->left != 0)
			changed |= orbits(p, follow[x->left - NB]);
		}
	    }
	}
    } while (changed);
}



/* Compute lookahead for each state */
void comp_lookaheads()
{
    int i;
    LR1 *x, *y;
    short *t;
    Gsym g, *s;
    bool changed;

    follow = talloc(sizeof(follow[0]) * nnonts);
    sbit(kernel[0][0].look, 0);
    do {
	changed = NO;
	for (i = 0; i < nstates; i++) {
	    comp_follow(i);
	    for (x = kernel[i]; x->item; x++) {
		if ((g = *(s = x->item)) != 0) {
		    s++;
		    for (t = action[i]; *t && thru[*t] != g; t++)
			;
		    assert(thru[*t] == g);
		    for (y = kernel[*t]; y->item && y->item != s; y++)
			;
		    assert(y->item == s);
		    changed |= orbits(y->look, x->look);
		}
	    }
	    for (t = action[i]; *t; t++) {
		for (x = kernel[*t]; x->item; x++) {
		    if (x->left != 0)
			changed |= orbits(x->look, follow[x->left - NB]);
		}
	    }
	    for (x = kernel[i]; x->item; x++) {
		if (x->item[0] == 0 && x->item[-2] == 0)
		    orbits(x->look, follow[x->item[-1] - NB]);
	    }
	}
    } while (changed);
    
    if (debug) {
	for (i = 0; i < nstates; i++) {
	    fprintf(vfp, "state %d:\n", i);
	    for (x = kernel[i]; x->item != NULL; x++) {
		write_item(x->item);
		fprintf(vfp, "\t\t[ ");
		dump_set(x->look);
		fprintf(vfp, "]\n");
	    }
	}
    }
}



/* Compare precedence of production (pnum) and grammar symbol x */
#define NON_ASSOC -32768
int cmpprec(pnum, x)
    int pnum;
    Gsym x;
{
    int v;
    
    if (gram[pnum]->assoc == A_UNDEF || (gsym[x]->assoc & A_MASK) == A_UNDEF)
	return (0);
    if ((v = (int)gsym[x]->prec - (int)gram[pnum]->prec) != 0)
	return (v);
    switch (gram[pnum]->assoc) {
    case A_LEFT:	/* Left associative, reduce */
	return (-1);
    case A_RIGHT:	/* Right associative, shift */
	return (1);
    case A_NON:		/* Non-associative, error */
	return (NON_ASSOC);
    }
}



/* Print transition table for state (st) */
void print_state(st)
    int st;
{
    LR1 *x;
    char *str;
    short *s;
    Reduce *r;

    if (!vfp)
	return;
    fprintf(vfp, "state %d\n", st);
    for (x = kernel[st]; x->item; x++)
	write_item(x->item);
    putc('\n', vfp);
    for (s = action[st], r = reduce[st]; ; ) {
	if (thru[*s] < r->sym) {
	    str = gsym[thru[*s]]->name;
	    fprintf(vfp, strlen(str) < 8 ? "\t%s\t\t" : "\t%s\t", str);
	    fprintf(vfp, "%s %d\n", isterm(thru[*s]) ? "shift" : "goto", *s);
	    s++;
	} else {
	    str = r->sym == NILSYM ? "." : gsym[r->sym]->name;
	    fprintf(vfp, strlen(str) < 8 ? "\t%s\t\t" : "\t%s\t", str);
	    if (r->num == 0)
		fprintf(vfp, "accept\n");
	    else if (r->num < 0)
		fprintf(vfp, "error\n");
	    else
		fprintf(vfp, "reduce (%d)\n", r->num);
	    if (*s == 0 && r->sym == NILSYM)
		break;
	    r++;
	}
    }
    putc('\n', vfp);
}



int cmp_by_num(x, y)
    Reduce *x, *y;
{
    if (x->num != y->num)
	return (y->num - x->num);
    return (x->sym - y->sym);
}


int cmp_by_sym(x, y)
    Reduce *x, *y;
{
    if (x->sym != y->sym)
	return (x->sym - y->sym);
    return (x->num - y->num);
}


/* Fill reduce entries for each state */
void fill_reduce()
{
    int i, j, k;
    LR1 *x;
    short *t, *s;
    Reduce tmpr[300];
    int nr;
    int tdefact, pnum;
    int rel;
    int maxn;
    Gsym e;
    uchar *alook = talloc(bytes);

    memset(tmpflag, 0, sizeof(bool) * MAXPROD);
    for (i = 0; i < nstates; i++) {
	tdefact = 0;
	for (t = action[i]; *t; t++) {
	    if (thru[*t] == 1) /* shifting error */
		tdefact = -1;
	}

	/* Pick up reduce entries */
	nr = 0;
	for (x = kernel[i]; x->item; x++) if (*x->item == 0) {
	    memcpy(alook, x->look, bytes);

	    pnum = item2pnum(x->item);
	    /* find shift/reduce conflict */
	    for (t = action[i]; *t; t++) {
		if (*t < 0) continue;
		e = thru[*t];
		if (!isterm(thru[*t])) break;
		if (tbit(alook, e)) {
		    rel = cmpprec(pnum, e);
		    if (rel == NON_ASSOC) {
			cbit(alook, e);
			*t = -1;
			if (nr >= height(tmpr))
			    die("Too many actions in one state");
			tmpr[nr].sym = e;
			tmpr[nr++].num = -1;
#if 0
if (vfp) fprintf(vfp, "%d: error overrides on %s\n", i, gsym[e]->name);
#endif
		    } else if (rel < 0) /* reduce */ {
			*t = -1;
#if 0
if (vfp) fprintf(vfp, "%d: reduce overrides on %s\n", i, gsym[e]->name);
#endif
		    }
		    else if (rel > 0) /* shift */ {
			cbit(alook, e);
		    } else if (rel == 0) /* conflict */ {
			cbit(alook, e);
			nsrerr++;
			if (vfp) {
			    fprintf(vfp,
		"%d: shift/reduce conflict (shift %d, reduce %d) on %s\n",
				    i, *t, pnum, gsym[e]->name);
			}
		    }
		}
	    }
	    for (j = 0; j < nr; j++) {
		if (tbit(alook, tmpr[j].sym)) {
		    /* reduce/reduce conflict */
		    nrrerr++;
		    if (vfp) {
			fprintf(vfp,
		"%d: reduce/reduce conflict (reduce %d, reduce %d) on %s\n",
				i, tmpr[j].num, pnum, gsym[tmpr[j].sym]->name);
		    }
		    if (pnum < tmpr[j].num)
			tmpr[j].num = pnum;
		    cbit(alook, tmpr[j].sym);
		}
	    }

	    for_each_member (e, alook) {
		if (nr >= height(tmpr))
		    die("Too many actions in one state");
		tmpr[nr].sym = e;
		tmpr[nr].num = pnum;
		nr++;
	    }
	}
	
	/* Decide default action */
	if (!tdefact) {
	    tdefact = -1;
	    qsort(tmpr, nr, sizeof(tmpr[0]), cmp_by_num);
	    maxn = 0;
	    for (j = 0; j < nr; ) {
		for (k = j; j < nr && tmpr[j].num == tmpr[k].num; j++)
		    ;
		if (j - k > maxn && tmpr[k].num > 0) {
		    maxn = j - k;
		    tdefact = tmpr[k].num;
		}
	    }
	}
	/* Squeeze tmpr */
	for (j = k = 0; j < nr; j++) {
	    if (tmpr[j].num != tdefact) {
		if (k != j)
		    tmpr[k] = tmpr[j];
		k++;
	    }
	}
	nr = k;
	qsort(tmpr, nr, sizeof(tmpr[0]), cmp_by_sym);
	tmpr[nr].sym = NILSYM;
	tmpr[nr++].num = tdefact;

	/* Squeeze shift actions */
	for (s = t = action[i]; *t; t++) {
	    if (*t >= 0)
		*s++ = *t;
	}
	*s = 0;

	for (j = 0; j < nr; j++) {
	    if (tmpr[j].num >= 0)
		tmpflag[tmpr[j].num] = 1;
	}
	/* Register tmpr */
	reduce[i] = alloc(sizeof(Reduce) * nr);
	nacts2 += nr;
	memcpy(reduce[i], tmpr, sizeof(tmpr[0]) * nr);
	
	/* print states */
	if (vfp)
	    print_state(i);
    }

    k = 0;
    for (j = 0; j < nprods; j++) {
	if (!tmpflag[j]) {
	    k++;
	    if (vfp) {
		fprintf(vfp, "Never reduced: ");
		write_item(gram[j]->body + 1);
	    }
	}
    }
    if (k) fprintf(stderr, "%d rule(s) never reduced\n", k);
}




global void comp_lalr()
{
    /* precompute symbols that reach empty production */
    comp_empty();

    /* precompute first set & nullable set */
    first_nullable_precomp();

    /* compute kernels of LR(0) items & fill shift entries */
    kernel = talloc(sizeof(kernel[0]) * MAXSTATE);
    comp_kernels();
#if 0
 fprintf(stderr, "kernel done. ");
 show_mem_usage();
#endif

    /* compute lookaheads */
    comp_lookaheads();
#if 0
 fprintf(stderr, "lookahead done. ");
 show_mem_usage();
#endif

    /* fill reduce entries & check conflictions */
    fill_reduce();
#if 0
 fprintf(stderr, "fill_reduce done. ");
 show_mem_usage();
#endif

    /* print diagnostic messages */
    if (nsrerr != expected_srconf || nrrerr != 0) {
	char *conj = "";
	fprintf(stderr, "%s: there are", filename);
	if (nsrerr != expected_srconf) {
	    fprintf(stderr, " %d shift/reduce", nsrerr);
	    conj = " and";
	}
	if (nrrerr != 0)
	    fprintf(stderr, "%s %d reduce/reduce", conj, nrrerr);
	fprintf(stderr, " conflicts\n");
    }

    if (vfp) {
	fprintf(vfp, "\nStatistics for %s:\n", filename);
	fprintf(vfp, "\t%d terminal symbols\n", nterms);
	fprintf(vfp, "\t%d nonterminal symbols\n", nnonts);
	fprintf(vfp, "\t%d productions\n", nprods);
	fprintf(vfp, "\t%d states\n", nstates);
	fprintf(vfp, "\t%d shift/reduce, %d reduce/reduce conflicts\n",
		nsrerr, nrrerr);
	fprintf(vfp, "\t%d items\n", nitems);
	fprintf(vfp, "\t%d lookahead sets used\n", nlooks);
	fprintf(vfp, "\t%d+%d=%d action entries\n", nacts, nacts2, nacts + nacts2);
	fprintf(vfp, "\t%ld bytes used\n", bytes_allocated());
    }

    release();
}
