/*
 * Table Optimizer
 */

#include "config.h"

#include <stdio.h>
#if HAS_STDLIB
# include <stdlib.h>
#endif /* HAS_STDLIB */
#include <assert.h>

#include "common.h"
#include "yacc.h"
#include "misc.h"
#include "grammar.h"
#include "lalr.h"
#include "table.h"


#define SCHAR_MAX 127


static int n2tabst;
static short *stateindex;
static short *indexstate;
static int nstcl;
static short *stcl;
static short *father;
static short *xrvec;
static int nxr;

static short *yynext;
static short *yygbase;
static short *yybase;
static short *yydefgoto;
static int yylast;
static int yyunused;

static int maxtokenval;


#define YYFLAG 32767
#define YYPASS 16384




/* Return non-zero if state i and state j have same shift/reduce actions */
int same_action(i, j)
    int i;
    int j;
{
    short *t, *s;
    Reduce *r, *q;

    if (i == j)
	return 1;
    for (r = reduce[i], q = reduce[j]; r->sym != NILSYM; r++, q++)
	if (r->sym != q->sym || r->num != q->num)
	    return 0;
    if (q->sym != NILSYM)
	return 0;
    for (t = action[i], s = action[j]; isterm(thru[*t]); t++, s++) {
	if (*t != *s)
	    return 0;
    }
    return !isterm(thru[*s]);
}


/* Return number of common actions between state i and j */
int common_action(i, j, dp, iop, jop)
    int i;
    int j;
    int *dp;
    int *iop;
    int *jop;
{
    short *t = action[i], *s = action[j];
    int comm = 0, diff = 0, ionly = 0, jonly = 0;

    while (isterm(thru[*t]) || isterm(thru[*s])) {
	if (*t == *s) {
	    t++;
	    s++;
	    comm++;
	} else if (thru[*t] < thru[*s]) {
	    t++;
	    ionly++;
	} else if (thru[*t] > thru[*s]) {
	    s++;
	    jonly++;
	} else {
	    t++;
	    s++;
	    diff++;
	}
    }
    if (dp) *dp = diff;
    if (iop) *iop = ionly;
    if (jop) *jop = jonly;
    return comm;
}



int diff_action(i, j, diffs)
    int i;
    int j;
    short *diffs;
{
    short *t = action[i], *s = action[j];
    int ndiff = 0;

    while (isterm(thru[*t]) || isterm(thru[*s])) {
	if (*t == *s) {
	    t++;
	    s++;
	} else if (thru[*t] < thru[*s]) {
	    diffs[ndiff++] = *t;
	    t++;
	} else if (thru[*t] > thru[*s]) {
	    diffs[ndiff++] = -*s;
	    s++;
	} else {
	    diffs[ndiff++] = *t;
	    t++;
	    s++;
	}
    }
    diffs[ndiff] = 0;
    return ndiff;
}



int actsize(i)
    int i;
{
    int n = 0;
    short *t;

    for (t = action[i]; isterm(thru[*t]); t++)
	n++;
    return n;
}


int cmp_actsize(x, y)
    short *x;
    short *y;
{
    int n = actsize(*x) - actsize(*y);
    if (n != 0)
	return n;
    return *x - *y;
}


void separate_table()
{
    int i, j;
    short *nsons;

    stcl = talloc(sizeof(short) * nstates);
    nstcl = 0;
    for (i = 0; i < nstates; i++) {
	for (j = 0; j < nstcl; j++) {
	    if (same_action(stcl[j], i))
		break;
	}
	if (j == nstcl)
	    stcl[nstcl++] = i;
    }
    qsort(stcl, nstcl, sizeof(stcl[0]), cmp_actsize);

    father = talloc(sizeof(short) * nstcl);
    nsons = talloc(sizeof(short) * nstcl);

    for (i = 0; i < nstcl; i++)
	nsons[i] = 1;
    for (i = 0; i < nstcl; i++) {
	int comm, diff, jonly, ionly;
	int m = 0;

	father[i] = i;
	for (j = i + 1; j < nstcl; j++) {
	    if (reduce[stcl[j]][0].sym != NILSYM)
		continue;
	    comm = common_action(stcl[i], stcl[j], &diff, &ionly, &jonly);
	    diff += ionly + jonly * nsons[i];
	    if (comm - diff > m) {
		m = comm - diff;
		father[i] = j;
		nsons[j] += nsons[i];
	    }
	}
    }
    for (i = 0; i < nstcl; i++) {
	for (j = i; father[j] != j; j = father[j])
	    ;
	if (j != i)
	    father[i] = j;
    }

}



void reg_extra_reduce()
{
    int rederr[100];
    int i, j, n, m;
    Reduce *r;

    m = 0;
    for (i = 0; i < nstcl; i++) {
	n = 0;
	for (r = reduce[stcl[i]]; r->sym != NILSYM; r++) {
	    for (j = 0; j < n; j++) {
		if (rederr[j] == r->num)
		    break;
	    }
	    if (j >= n)
		rederr[n++] = r->num;
	}
	m += n;
    }

    xrvec = talloc(sizeof(short) * 2 * m);
    nxr = m;

    n = 0;
    for (i = 0; i < nstcl; i++) {
	m = n;
	for (r = reduce[stcl[i]]; r->sym != NILSYM; r++) {
	    for (j = m; j < n; j+=2) {
		if (xrvec[j] == r->num)
		    break;
	    }
	    if (j >= n) {
		xrvec[n++] = r->num;
		xrvec[n++] = i;
	    }
	}
    }

}




struct gotopair {
    short from;
    short to;
};

int cmp_gotopair(x, y)
    struct gotopair *x, *y;
{
    return (x->to - y->to);
}

struct order {
    short num;
    short span;
    short index;
};

int cmp_order(x, y)
    struct order *x, *y;
{
#if 0
    if (y->span != x->span)
	return (y->span - x->span);
    if (y->num != x->num)
	return (y->num - x->num);
#endif

#if 1
    if (y->num != x->num)
	return (y->num - x->num);
    if (y->span != x->span)
	return (y->span - x->span);
#endif
    return ((unsigned)x->index - (unsigned)y->index);
}


static struct order *order;

/* Compute allocation order */
void decide_order()
{
    struct gotopair gotmp[200];
    int i, m, k, g;
    short *t;

    order = talloc(sizeof(struct order) * (nstcl + nnonts));

    /* shift/reduce table size and span */
    for (i = 0; i < nstcl; i++) {
	int minsym = 10000;
	int maxsym = -1;
	int n = 0;
	int v;
	short diffs[300];
	Reduce *r;

	if (father[i] != i) {
	    diff_action(stcl[i], stcl[father[i]], diffs);
	    t = diffs;
	} else {
	    t = action[stcl[i]];
	}
	for (r = reduce[stcl[i]]; r->sym != NILSYM; r++) {
	    v = gsym[r->sym]->value;
	    if (v < minsym) minsym = v;
	    if (v > maxsym) maxsym = v;
	    n++;
	}
	for (; isterm(thru[abs(*t)]); t++) {
	    v = gsym[thru[abs(*t)]]->value;
	    if (v < minsym) minsym = v;
	    if (v > maxsym) maxsym = v;
	    n++;
	}
	order[i].span = maxsym < 0 ? 0 : maxsym - minsym + 1;
	order[i].num = n;
	order[i].index = i;
    }

    /* Count GOTO Table entries for nonterminal g */
    for (g = NB + 1; g < NB + nnonts; g++) {
	int maxfrom, minfrom;
	int n = 0;
	for (i = 0; i < nstates; i++) {
	    for (t = action[i]; *t; t++) {
		if (thru[*t] == g) {
		    gotmp[n].from = stateindex[i];
		    gotmp[n++].to = stateindex[*t];
		    break;
		}
	    }
	}
	qsort(gotmp, n, sizeof(gotmp[0]), cmp_gotopair);

	/* Find default (most often) entry */
	m = 0;
	yydefgoto[g - NB] = YYFLAG;
	for (i = 0; i < n; ) {
	    for (k = i; i < n && gotmp[i].to == gotmp[k].to; i++)
		;
	    if (i - k > m) {
		m = i - k;
		yydefgoto[g - NB] = gotmp[k].to;
	    }
	}
	maxfrom = -1;
	minfrom = 10000;
	for (i = 0; i < n; i++) {
	    if (gotmp[i].to != yydefgoto[g - NB]) {
		if (gotmp[i].from > maxfrom)
		    maxfrom = gotmp[i].from;
		if (gotmp[i].from < minfrom)
		    minfrom = gotmp[i].from;
	    }
	}
	order[nstcl + g - (NB + 1)].span = maxfrom < 0 ? 0 : maxfrom - minfrom + 1;
	order[nstcl + g - (NB + 1)].num = n - m;
	order[nstcl + g - (NB + 1)].index = -g;
    }
    qsort(order, nstcl + nnonts - 1, sizeof(order[0]), cmp_order);
    if (debug) {
	fprintf(vfp, "table assignment order:\n");
	for (i = 0; i < nstcl + nnonts - 1; i++) {
	    if (order[i].index >= 0)
		fprintf(vfp, "\tstate %d\n", stcl[order[i].index]);
	    else
		fprintf(vfp, "\tgoto %s\n", gsym[-order[i].index]->name);
	}
    }
}


/* Create table */
void maketable()
{
    struct gotopair gotmp[200];
    int i, j, k, n, lfree, cl, g, smallest;

    /* Initialize */
    for (i = 0; i < MAXTABLE; i++)
	yynext[i] = YYFLAG;
    for (i = 0; i < nstates; i++)
	yybase[i] = YYFLAG;
    for (i = 0; i < nnonts; i++)
	yygbase[i] = YYFLAG;
    lfree = 0;

    for (i = 0; i < nstcl + nnonts - 1; i++) {
	if ((cl = order[i].index) >= 0) {
	    short diffs[300];
	    short *v, *t;
	    Reduce *r = reduce[stcl[cl]];

	  /* Fill shift/reduce entries */
	    if (father[cl] != cl) {
		diff_action(stcl[cl], stcl[father[cl]], diffs);
		v = diffs;
	    } else {
		v = action[stcl[cl]];
	    }
	    if (!isterm(thru[abs(*v)]) && r->sym == NILSYM)
		continue;
	    smallest = maxtokenval;
	    for (; r->sym != NILSYM; r++) {
		if (gsym[r->sym]->value < smallest)
		    smallest = gsym[r->sym]->value;
	    }
	    for (t = v; isterm(thru[abs(*t)]); t++) {
		if (gsym[thru[abs(*t)]]->value < smallest)
		    smallest = gsym[thru[abs(*t)]]->value;
	    }
	    /* Find room in yynext */
	    for (k = lfree - smallest; k < MAXTABLE; k++) {
		for (r = reduce[stcl[cl]]; r->sym != NILSYM; r++) {
		    n = k + gsym[r->sym]->value;
		    if (n >= MAXTABLE)
			die("Table overflow");
		    if (yynext[n] != YYFLAG)
			goto skip;
		}
		for (t = v; isterm(thru[abs(*t)]); t++) {
		    n = k + gsym[thru[abs(*t)]]->value;
		    if (n >= MAXTABLE)
			die("Table overflow");
		    if (yynext[n] != YYFLAG)
			goto skip;
		}
		for (j = 0; j < i; j++) {
		    if (order[j].index >= 0
			&& yybase[stateindex[stcl[order[j].index]]] == k)
			goto skip;
		}
		break;
	    skip: ;
	    }
	    if (k >= MAXTABLE)
		die("Table overflow");
	    yybase[stateindex[stcl[cl]]] = k;
	    for (t = v; isterm(thru[abs(*t)]); t++) {
		yynext[k + gsym[thru[abs(*t)]]->value]
		    = *t >= 0 ? stateindex[*t]
			: YYPASS + gsym[thru[-*t]]->value;
	    }
	    for (r = reduce[stcl[cl]]; r->sym != NILSYM; r++) {
		for (j = 0; xrvec[j] != r->num || xrvec[j+1] != cl; j+=2)
		    ;
		yynext[k + gsym[r->sym]->value] = nstates + j / 2;
	    }
	} else {
	    /* Fill goto entries */
	    g = -cl;
	    n = 0;
	    smallest = nstates;
	    for (j = 0; j < nstates; j++) {
		short *t;

		for (t = action[j]; *t; t++) {
		    if (thru[*t] == g && stateindex[*t] != yydefgoto[g - NB]) {
			gotmp[n].from = stateindex[j];
			if (stateindex[j] < smallest)
			    smallest = stateindex[j];
			gotmp[n++].to = stateindex[*t];
			break;
		    }
		}
	    }
	    if (n == 0)
		continue;
	    for (k = lfree - smallest; k < MAXTABLE; k++) {
		for (j = 0; j < n; j++) {
		    int p = k + gotmp[j].from;
		    if (p >= MAXTABLE)
			die("Table overflow");
		    if (yynext[p] != YYFLAG)
			goto skip2;
		}
		for (j = 0; j < i; j++) {
		    if (order[j].index < 0
			&& yygbase[-order[j].index - NB] == k)
			goto skip2;
		}
		break;
	    skip2: ;
	    }
	    if (k >= MAXTABLE)
		die("Table overflow");
	    yygbase[g - NB] = k;
	    for (j = 0; j < n; j++) {
		yynext[k + gotmp[j].from] = gotmp[j].to;
	    }
	}

	/* maintain lfree */
	while (yynext[lfree] != YYFLAG)
	    lfree++;
    }

    for (i = 0; i < nstates; i++) {
	if (yybase[stateindex[i]] == YYFLAG
	    && (isterm(thru[action[i][0]]) || reduce[i][0].sym != NILSYM)) {
	    for (j = 0; j < nstcl; j++) {
		if (same_action(stcl[j], i)) {
		    yybase[stateindex[i]] = yybase[stateindex[stcl[j]]];
		    break;
		}
	    }
	}
    }

    /* Count table size and unused entries  */
    yylast = 0;
    yyunused = 0;
    n = 0;
    for (i = 0; i < MAXTABLE; i++) {
	if (yynext[i] != YYFLAG) {
	    yylast = i + 1;
	    yyunused = n;
	} else {
	    n++;
	}
    }
}


/* Return length of production p */
int length(p)
    Gsym *p;
{
    int n;
    
    for (n = 0; *++p; n++)
	;
    return (n);
}


#if 0
/* Quote string s and print */
void qputs(s, fp)
    char *s;
    FILE *fp;
{
    char *p;
    
    p = s;
    if (s[0] == '\'')
	p++;
    for (; *p; p++) {
	if (s[0] == '\'' && p[1] == '\0')
	    break;
	if (*p == '\\' || *p == '"')
	    putc('\\', fp);
	putc(*p, fp);
    }
}
#endif


void renumber_states()
{
    int i, j, k, n, m;

    stateindex = talloc(sizeof(short) * nstates);
    n = 0;
    m = nstates;
    for (i = 0; i < nstates; i++) {
	if (!isterm(thru[action[i][0]]) && reduce[i][0].sym == NILSYM) {
	    --m;
	} else {
	    for (j = 0; j < nstcl; j++) {
		if (same_action(stcl[j], i))
		    break;
	    }
	    if (father[j] != j)
		n++;
	}
    }
    n2tabst = n;

    if (vfp) {
	fprintf(vfp, "\t%d LR(0) states\n", nstates - m);
	fprintf(vfp, "\t%d 2-table states\n", n);
    }

    k = 0;
    for (i = 0; i < nstates; i++) {
	if (!isterm(thru[action[i][0]]) && reduce[i][0].sym == NILSYM) {
	    stateindex[i] = m++;
	} else {
	    for (j = 0; j < nstcl; j++) {
		if (same_action(stcl[j], i))
		    break;
	    }
	    if (father[j] != j) {
		stateindex[i] = k++;
	    } else {
		stateindex[i] = n++;
	    }
	}
    }
    assert(k == n2tabst);
    indexstate = talloc(sizeof(short) * nstates);
    for (i = 0; i < nstates; i++)
	indexstate[stateindex[i]] = i;
}



void gentab2();


/* Generate table */
global void gentable()
{
    int i, j, n, m;

    separate_table();

    reg_extra_reduce();

    renumber_states();

    yynext = talloc(sizeof(short) * MAXTABLE);
    yygbase = talloc(sizeof(short) * nnonts);
    yybase = talloc(sizeof(short) * nstates);
    yydefgoto = talloc(sizeof(short) * nnonts);

    maxtokenval = 0;
    for (i = 1; i < nterms; i++) {
    	if (gsym[i]->value > maxtokenval)
	    maxtokenval = gsym[i]->value;
    }

    decide_order();

    maketable();
    
    fprintf(ofp, "\n#ifdef YYDEBUG\n");
#if 0
    /* Generate production names */
    fprintf(ofp, "static char *yypnames[] = {\n");
    for (i = 0; i < nprods; i++) {
	Gsym *q;
	fprintf(ofp, "  \"%s :", gsym[gram[i]->body[1]]->name);
	for (q = gram[i]->body + 2; *q != 0; q++) {
	    putc(' ', ofp);
	    qputs(gsym[*q]->name, ofp);
	}
	fprintf(ofp, "\",\n");
    }
    fprintf(ofp, "};\n");
#endif
    
    /* Generate terminal symbols' names */
    m = 0; n = maxtokenval;
    for (i = 1; i < nterms; i++) {
	if (gsym[i]->name[0] != '\'') {
	    if (gsym[i]->value > m)
		m = gsym[i]->value;
	    if (gsym[i]->value < n)
		n = gsym[i]->value;
	}
    }
    fprintf(ofp, "#define YYMINTOK %d\n", n);
    fprintf(ofp, "#define YYMAXTOK %d\n", m);
    fprintf(ofp, "static char *yytnames[] = {\n");
    for (i = n; i <= m; i++) {
	for (j = 1; j < nterms; j++) {
	    if (gsym[j]->value == i) {
		fprintf(ofp, " \"%s\",\n", gsym[j]->name);
		break;
	    }
	}
	if (j == nterms)
	    fprintf(ofp, "  0,\n");
    }
    fprintf(ofp, "  0 };\n\n");

    /* Generate nonterminal symbols' names */
    fprintf(ofp, "static char *yyntnames[] = {\n");
    for (i = 0; i < nnonts; i++) {
	fprintf(ofp, "  \"%s\",\n", gsym[NB + i]->name);
    }
    fprintf(ofp, "};\n\n");

    /* Generate real state numbers */
    fprintf(ofp, "static short yyrstates[] = {");
    for (i = 0; i < nstates; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", indexstate[i]);
    }
    fprintf(ofp, "\n};\n#endif /* YYDEBUG */\n\n");
    
    /* Generate yynext */
    fprintf(ofp, "#define YYPASS %d\n", YYPASS);
    fprintf(ofp, "#define YYLAST %d\n", yylast);
    fprintf(ofp, "static short yynext[] = {");
    for (i = 0; i < yylast; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", yynext[i]);
    }
    fprintf(ofp, "\n};\n\n");
    
    /* Generate yybase */
    for (n = nstates; --n >= 0 && yybase[n] == YYFLAG; )
	;
    n++;
    fprintf(ofp, "#define YY2TABSTATES %d\n", n2tabst);
    fprintf(ofp, "#define YYLR1STATES %d\n", n);
    fprintf(ofp, "#define YYNSTATES %d\n", nstates);
    fprintf(ofp, "#define YYNXR %d\n", nstates + nxr);
    fprintf(ofp, "#define YYSTATE0 %d\n", stateindex[0]);
    fprintf(ofp, "static short yybase[] = {");
    for (i = 0; i < n; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", yybase[i]);
    }
    fprintf(ofp, "\n};\n\n");

    fprintf(ofp, "static short yybase2[] = {");
    for (i = 0; i < n2tabst; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	for (j = 0; j < nstcl; j++) {
	    if (same_action(stcl[j], indexstate[i]))
		break;
	}
	assert(reduce[stcl[father[j]]][0].sym == NILSYM);
	fprintf(ofp, "%6d,", yybase[stateindex[stcl[father[j]]]]);
    }
    fprintf(ofp, n2tabst ? "\n};\n\n" : " 0\n};\n\n");

    gentab2();
}


void gentab2()
{
    int i, j;
    char *type;

    /* Generate yydefault */
    fprintf(ofp, "static short yydefault[] = {");
    j = 0;
    for (i = 0; i < nstates; i++) {
	Reduce *r = reduce[indexstate[i]];
	for (; r->sym != NILSYM; r++);
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", r->num);
    }
    fprintf(ofp, "\n};\n\n");
    
    /* Generate yycheck */
    fprintf(ofp, "static short yycheck[] = {");
    for (i = 0; i < nstates; i++) {
	j = indexstate[i];
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", j == 0 ? YYFLAG :
			     isterm(thru[j]) ? gsym[thru[j]]->value :
			                       -(thru[j] - NB));
    }
    for (i = 0; i < nxr; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", yybase[stateindex[stcl[xrvec[2*i+1]]]]);
    }
    fprintf(ofp, "\n};\n\n");

    /* Generate yyrederr */
    fprintf(ofp, "static short yyrederr[] = {");
    for (i = 0; i < nxr; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", xrvec[2*i]);
    }
    fprintf(ofp, "\n};\n\n");
	
    /* Generate yygbase */
    fprintf(ofp, "static short yygbase[] = {");
    for (i = 0; i < nnonts; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", yygbase[i]);
    }
    fprintf(ofp, "\n};\n\n");

    /* Generate yydefgoto */
    fprintf(ofp, "static short yydefgoto[] = {");
    for (i = 0; i < nnonts; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", yydefgoto[i]);
    }
    fprintf(ofp, "\n};\n\n");
    
    /* Generate yyleft */
    fprintf(ofp, "static %s yyleft[] = {",
	    nnonts <= SCHAR_MAX ? "char" : "short");
    for (i = 0; i < nprods; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", gram[i]->body[1] - NB);
    }
    fprintf(ofp, "\n};\n\n");
    
    /* Generate yylen */
    type = "char";
    for (i = 0; i < nprods; i++) {
	if (length(gram[i]->body + 1) > SCHAR_MAX) {
	    type = "short";
	    break;
	}
    }
    fprintf(ofp, "static %s yylen[] = {", type);
    for (i = 0; i < nprods; i++) {
	if (i % 10 == 0)
	    fprintf(ofp, "\n");
	fprintf(ofp, "%6d,", length(gram[i]->body + 1));
    }
    fprintf(ofp, "\n};\n\n");

    if (vfp) {
	fprintf(vfp, "\t%d shift transition classes\n", nstcl);
	fprintf(vfp, "\t%d table entries, %d unused\n", yylast, yyunused);
	fprintf(vfp, "\t%ld bytes used by table optimizer\n", bytes_allocated());
    }
    
}
